# validator.py
class Validator:
    @staticmethod
    def validate_args(args_needed: list, args_user: dict):
        """
        Valida se o dicionário fornecido contém os parâmetros necessários.\n
        Validates if the provided dictionary contains the required parameters.

        Args:
            args_needed (list): Lista de chaves de configuração necessárias.
                                List of required configuration keys.
            args_user (dict): Dicionário de configuração do usuário.
                              User's configuration dictionary.

        Raises:
            KeyError: Se um parâmetro necessário estiver faltando no dicionário do usuário.
                      If a required parameter is missing in the user's dictionary.

        Examples:
            >>> Validator.validate_args(['table_checkpoint_location'], {'table_checkpoint_location':'/save/_checkpointLocation'})
        """
        for arg in args_needed:
            if arg not in args_user:
                raise KeyError(
                    f'Nao foi possivel localizar o parametro: {arg} . Por favor adicionar'
                    # Could not find the parameter: {arg}. Please add it.
                )



# config_ingestor.py
from carlton.ingest.config_validator import ConfigValidator
from carlton.ingest.path_builder import PathBuilder
from carlton.utils.logger import log_error, log_info


class ConfigIngestor:
    @staticmethod
    def get_params_path(config):
        """
        Obtém os caminhos apropriados com base no tipo de recurso do arquivo.\n
        Get the appropriate paths based on the file resource type.

        Args:
            config (dict): Dicionário de configuração contendo as chaves necessárias.
                           Configuration dictionary containing necessary keys.

        Returns:
            dict: Dicionário de configuração atualizado com os caminhos.
                  Updated configuration dictionary with paths.

        Example:
            >>> config = {'file_resource': 'adls','container_src': 'source_container','storage_name_src': 'source_storage', 'container_tgt': 'target_container','storage_name_tgt': 'target_storage','path_src': 'source_path','table_name': 'name_table'}
            >>> updated_config = ConfigIngestor.get_params_path(config)
        """
        if config['file_resource'] == 'adls':
            ConfigValidator.validate_args(
                [
                    'container_src',
                    'storage_name_src',
                    'container_tgt',
                    'storage_name_tgt',
                    'path_src',
                    'table_name',
                ],
                config,
            )
            log_info('Building paths for ADLS')
            config = PathBuilder.build_adls_paths(config)
        return config

    @staticmethod
    def config_ingest_src(config, custom_config_spark={}):
        """
        Configura as configurações de ingestão da fonte.\n
        Configure the source ingestion settings.

        Args:
            config (dict): Dicionário de configuração contendo as chaves necessárias.
                           Configuration dictionary containing necessary keys.
            custom_config_spark (dict, optional): Dicionário de configuração personalizada do Spark.
                                                  Custom Spark configuration dictionary.

        Returns:
            dict: Dicionário com a configuração do autoloader.
                  Dictionary with autoloader configuration.

        Example:
            >>> config = {'type_run': 'batch','file_extension': 'csv','file_resource': 'adls','container_src': 'source_container','storage_name_src': 'source_storage','container_tgt': 'target_container','storage_name_tgt': 'target_storage','path_src': 'source_path','file_header': 'true','file_delimiter': ',','schemaLocation': 'schema_location','table_name': 'name_table'}
            >>> autoloader_config = ConfigIngestor.config_ingest_src(config)
        """
        ConfigValidator.validate_args(
            ['type_run', 'file_extension', 'file_resource'], config
        )

        log_info(
            'prepara os caminhos apropriados com base no tipo de recurso do arquivo'
        )
        config = ConfigIngestor.get_params_path(config)

        autoloader_config_csv = {}
        if config['file_extension'] == 'csv':
            ConfigValidator.validate_args(
                ['file_header', 'file_delimiter'], config
            )
            autoloader_config_csv = {
                'header': config['file_header'],
                'delimiter': config['file_delimiter'],
            }

        autoloader_config_json = {}
        if config['file_extension'] == 'json':
            autoloader_config_json = {
                'multiLine': True,
            }

        autoloader_config = {
            'pathGlobfilter': f"*.{config['file_extension']}",
            'cloudFiles.format': config['file_extension'],
            'cloudFiles.schemaLocation': config['schemaLocation'],
            'cloudFiles.schemaEvolutionMode': 'rescue',
            'cloudFiles.inferColumnTypes': 'false',
            'cloudFiles.allowOverwrites': 'true',
            'rescuedDataColumn': 'carlton_rescued',
            **autoloader_config_csv,
            **autoloader_config_json,
            **custom_config_spark,
        }

        return autoloader_config

    @staticmethod
    def config_ingest_tgt(config, custom_config_spark={}):
        """
        Configura as configurações de ingestão do destino.\n
        Configure the target ingestion settings.

        Args:
            config (dict): Dicionário de configuração contendo as chaves necessárias.
                           Configuration dictionary containing necessary keys.
            custom_config_spark (dict, optional): Dicionário de configuração personalizada do Spark.
                                                  Custom Spark configuration dictionary.

        Returns:
            dict: Dicionário com a configuração de salvamento.
                  Dictionary with save configuration.

        Example:
            >>> config = {'table_path': '/path/to/table','checkpointLocation': '/path/to/checkpoint'}
            >>> save_config = ConfigIngestor.config_ingest_tgt(config)
        """
        ConfigValidator.validate_args(
            ['table_path', 'checkpointLocation'], config
        )

        save_config = {
            'checkpointLocation': config['checkpointLocation'],
            'path': config['table_path'],
            'mergeSchema': True,
            **custom_config_spark,
        }

        return save_config



# data_saver.py
from pyspark.sql import DataFrame, SparkSession

from carlton.ingest.config_ingestor import ConfigIngestor
from carlton.ingest.config_validator import ConfigValidator
from carlton.utils.logger import log_error, log_info


class DataSaver:
    @staticmethod
    def save_data(
        spark: SparkSession,
        df: DataFrame,
        config_ingest: dict,
        custom_config_spark={},
    ) -> None:
        """
        Salva os dados do DataFrame em um destino especificado usando as configurações fornecidas.
        Saves the DataFrame data to a specified destination using the provided configurations.

        Args:
            df (DataFrame): DataFrame contendo os dados a serem salvos.
                            DataFrame containing the data to be saved.
            config_ingest (dict): Configurações de ingestão.
                                  Ingestion configurations.
            custom_config_spark (dict, optional): Configurações personalizadas do Spark.
                                                  Custom Spark configurations.

        Example:
            >>> from pyspark.sql import SparkSession
            >>> from data_saver import DataSaver
            >>> spark = SparkSession.builder.appName("example").getOrCreate()
            >>> data = [("Alice", 1), ("Bob", 2)]
            >>> columns = ["name", "value"]
            >>> df = spark.createDataFrame(data, columns)
            >>> config_ingest = {'schema_name': 'default','table_name': 'example_table','table_path': '/path/to/table','type_run': 'batch','trigger_processing_time': '10 seconds'}
            >>> custom_config_spark = {'spark.some.config.option': 'some-value'}
            >>> DataSaver.save_data(df, config_ingest, custom_config_spark)
        """
        try:
            # Valida os argumentos necessários
            # Validate the necessary arguments
            ConfigValidator.validate_args(
                ['schema_name', 'table_name', 'table_path', 'type_run'],
                config_ingest,
            )

            # Configura o destino com as configurações de ingestão
            # Configure the destination with the ingestion configurations
            save_config = ConfigIngestor.config_ingest_tgt(
                config_ingest, custom_config_spark
            )

            # Registra as configurações usadas para escrita
            # Log the configurations used for writing
            log_info('Configurations used for writing: ', msg_dict=save_config)

            # Lista de colunas internas que não devem ser incluídas na tabela
            # List of internal columns that should not be included in the table
            lst_builtin = [
                '_rescued',
                'carlton_current_date',
                'carlton_metadata',
            ]
            columns_file = ', '.join(
                f'{col.lower()} STRING'
                for col in df.columns
                if col not in lst_builtin
            )

            # Registra informações sobre a tabela sendo criada
            # Log information about the table being created
            log_info(
                f"Registering table {config_ingest['table_name']} in schema {config_ingest['schema_name']}"
            )
            log_info(
                f'Using liquid_cluster. Managed by the column carlton_current_date'
            )

            # Cria a tabela no destino especificado
            # Create the table in the specified destination
            query_create_table = f"""CREATE TABLE IF NOT EXISTS {config_ingest['schema_name']}.{config_ingest['table_name']} ({columns_file},_rescued STRING,carlton_current_date DATE,carlton_metadata struct<file_path:string,file_name:string,file_size:bigint,file_block_start:bigint,file_block_length:bigint,file_modification_time:timestamp>) USING DELTA CLUSTER BY (carlton_current_date) LOCATION '{config_ingest['table_path']}'"""

            log_info(f'Registering table: {query_create_table}')
            spark.sql(query_create_table)
            log_info(
                f"Table {config_ingest['schema_name']}.{config_ingest['table_name']} created successfully"
            )

            # Configura o tipo de trigger para escrita dos dados
            # Configure the trigger type for writing the data
            type_trigger = {}
            if config_ingest['type_run'] == 'batch':
                log_info(f'Identifying execution as batch')
                type_trigger['availableNow'] = True
            else:
                log_info(f'Identifying execution as stream')
                ConfigValidator.validate_args(
                    ['trigger_processing_time'], config_ingest
                )
                type_trigger['processingTime'] = config_ingest[
                    'trigger_processing_time'
                ]

            log_info(f'Starting record writing')

            # Escreve os dados no destino especificado
            # Write the data to the specified destination
            df.writeStream.options(**save_config).outputMode('append').trigger(
                **type_trigger
            ).table(
                f"{config_ingest['schema_name']}.{config_ingest['table_name']}"
            ).awaitTermination()

            log_info(
                f"{config_ingest['schema_name']}.{config_ingest['table_name']} processed successfully"
            )

        except Exception as e:
            # Registra qualquer erro que ocorra durante a escrita
            # Log any error that occurs during writing
            log_error(str(e))



import sys

from pyspark.sql import SparkSession

from carlton.ingest.table import read, save
from carlton.utils.logger import log_error, log_info
from carlton.utils.spark_session_manager import SparkSessionManager


def process_args(args):
    """
    Processa os argumentos passados para a função.\n
    Processes the arguments passed to the function.

    Args:
        args (list): Lista de argumentos.
                     List of arguments.

    Returns:
        dict: Dicionário de propriedades raiz.
              Dictionary of root properties.
    """
    root_properties = {}
    for i, arg in enumerate(args):
        if arg.startswith('-'):
            root_properties[arg.replace('-', '')] = args[i + 1]
    return root_properties


def main(args=sys.argv[1:]):
    """
    Função principal para executar a ingestão de dados.\n
    Main function to execute data ingestion.

    Args:
        args (list): Lista de argumentos passados para a função.
                     List of arguments passed to the function.

    Returns:
        None
    """
    try:
        log_info('Ingestão iniciada')
        # Data ingestion started

        # Processa os argumentos
        # Process the arguments
        print(args)
        root_properties = process_args(args)

        # Imprime as propriedades raiz
        # Print the root properties
        for p in root_properties:
            log_info(f'{p}: {root_properties[p]}')

        # Criação de SparkSession
        # Create SparkSession
        spark = SparkSessionManager.create_spark_session('Carlton Ingest APP')

        # Leitura de dados
        # Read data

        df = read(spark, root_properties)

        # Salvamento de dados
        # Save data
        save(spark, df, root_properties)

        log_info('Ingestão finalizada')
        # Data ingestion finished

    except Exception as e:
        # Loga qualquer erro que ocorrer
        # Log any error that occurs
        log_error(str(e))


if __name__ == '__main__':
    main()



# spark_session_manager.py
from pyspark.sql import SparkSession


class SparkSessionManager:
    @staticmethod
    def create_spark_session(app_name):
        """
        Cria ou obtém uma sessão Spark.
        Creates or gets a Spark session.

        Args:
            app_name (str): Nome do aplicativo Spark.
                            Spark application name.

        Returns:
            SparkSession: A sessão Spark.
                        The Spark session.
        """
        return SparkSession.builder.appName(app_name).getOrCreate()



# data_reader.py
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col, current_date

from carlton.ingest.config_ingestor import ConfigIngestor
from carlton.utils.logger import log_error, log_info


class DataReader:
    @staticmethod
    def read_data(
        spark: SparkSession, config_ingest: dict, custom_config_spark={}
    ) -> DataFrame:
        """
        Lê dados de uma fonte especificada usando as configurações fornecidas.
        Reads data from a specified source using the provided configurations.

        Args:
            config_ingest (dict): Configurações de ingestão.
                                  Ingestion configurations.
            custom_config_spark (dict, optional): Configurações personalizadas do Spark.
                                                  Custom Spark configurations.

        Returns:
            DataFrame: DataFrame contendo os dados lidos.
                       DataFrame containing the read data.

        Example:
            >>> config_ingest = {'type_run': 'batch','file_extension': 'csv','file_resource': 'adls','container_src': 'source_container','storage_name_src': 'source_storage','container_tgt': 'target_container','storage_name_tgt': 'target_storage','path_src': 'source_path','file_header': 'true','file_delimiter': ',','schemaLocation': 'schema_location','carlton_file_path': 'path/to/data'}
            >>> custom_config_spark = {'spark.some.config.option': 'some-value'}
            >>> df = DataReader.read_data(config_ingest, custom_config_spark)
        """
        try:
            # Configura o autoloader com as configurações de ingestão
            # Configure the autoloader with the ingestion configurations
            log_info('Configura o autoloader com as configurações de ingestão')
            autoloader_config = ConfigIngestor.config_ingest_src(
                config_ingest, custom_config_spark
            )

            # Registra as configurações usadas para leitura
            # Log the configurations used for reading
            log_info(f'Configurations used for reading: {autoloader_config}')

            log_info(f"load file from {config_ingest['carlton_file_path']}")

            # Lê os dados da fonte especificada e retorna um DataFrame
            # Read the data from the specified source and return a DataFrame
            return (
                spark.readStream.format('cloudFiles')
                .options(**autoloader_config)
                .load(config_ingest['carlton_file_path'])
                .select(
                    '*',
                    current_date().alias('carlton_current_date'),
                    col('_metadata').alias('carlton_metadata'),
                )
            )

        except Exception as e:
            # Registra qualquer erro que ocorra durante a leitura
            # Log any error that occurs during reading
            log_error(str(e))
            return None



# table.py
from pyspark.sql import DataFrame, SparkSession

from carlton.ingest.data_reader import DataReader
from carlton.ingest.data_saver import DataSaver
from carlton.utils.logger import log_info


def read(
    spark: SparkSession, config_ingest: dict, custom_config_spark={}
) -> DataFrame:
    """
    Função wrapper para ler dados usando DataReader.\n
    Wrapper function to read data using DataReader.

    Args:
        config_ingest (dict): Configurações de ingestão.
                              Ingestion configurations.
        custom_config_spark (dict, optional): Configurações personalizadas do Spark.
                                              Custom Spark configurations.

    Returns:
        DataFrame: DataFrame contendo os dados lidos.
                   DataFrame containing the read data.
    """
    log_info('DataReader.read_data')
    return DataReader.read_data(spark, config_ingest, custom_config_spark)


def save(
    spark: SparkSession,
    df: DataFrame,
    config_ingest: dict,
    custom_config_spark={},
) -> None:
    """
    Função wrapper para salvar dados usando DataSaver.\n
    Wrapper function to save data using DataSaver.

    Args:
        df (DataFrame): DataFrame contendo os dados a serem salvos.
                        DataFrame containing the data to be saved.
        config_ingest (dict): Configurações de ingestão.
                              Ingestion configurations.
        custom_config_spark (dict, optional): Configurações personalizadas do Spark.
                                              Custom Spark configurations.
    """
    log_info('DataSaver.save_data')
    DataSaver.save_data(spark, df, config_ingest, custom_config_spark)



# Databricks notebook source
import sys

from pyspark.sql import SparkSession

from carlton.ingest.table import read, save
from carlton.utils.logger import log_error, log_info
from carlton.utils.spark_session_manager import SparkSessionManager

# COMMAND ----------


def process_args(args):
    """
    Processa os argumentos passados para a função.\n
    Processes the arguments passed to the function.

    Args:
        args (list): Lista de argumentos.
                     List of arguments.

    Returns:
        dict: Dicionário de propriedades raiz.
              Dictionary of root properties.
    """
    root_properties = {}
    for i, arg in enumerate(args):
        if arg.startswith('-'):
            root_properties[arg.replace('-', '')] = args[i + 1]
    return root_properties


# COMMAND ----------

log_info('Ingestão iniciada')
# Data ingestion started

# Processa os argumentos
# Process the arguments
args = [
    'ingest',
    '-function',
    'ingest',
    '-storage_name_src',
    'stadrisk',
    '-container_src',
    'ctrdriskraw',
    '-file_resource',
    'adls',
    '-type_run',
    'batch',
    '-storage_name_tgt',
    'stadrisk',
    '-container_tgt',
    'dtmaster-catalog',
    '-schema_name',
    'bronze',
    '-table_name',
    'table_test',
    '-file_extension',
    'csv',
    '-path_src',
    'table_test',
    '-file_header',
    'true',
    '-file_delimiter',
    ',',
]
print(args)
root_properties = process_args(args)

# Imprime as propriedades raiz
# Print the root properties
for p in root_properties:
    log_info(f'{p}: {root_properties[p]}')

# Criação de SparkSession
# Create SparkSession
spark = SparkSessionManager.create_spark_session('Carlton Ingest APP')

# COMMAND ----------

df = read(spark, root_properties)



# config_validator.py
from carlton.utils.logger import log_info
from carlton.utils.validator import Validator


class ConfigValidator:
    @staticmethod
    def validate_args(required_args, config):
        """
        Validate that all required arguments are present in the config.

        :param required_args: List of required argument keys.
        :param config: Configuration dictionary to validate.
        """

        log_info(f'Validating configuration arguments: {required_args}')
        Validator.validate_args(required_args, config)



# path_builder.py
class PathBuilder:
    @staticmethod
    def build_adls_paths(config):
        """
        Build paths for Azure Data Lake Storage (ADLS).

        :param config: Configuration dictionary containing necessary keys.
        :return: Updated configuration dictionary with ADLS paths.
        """
        config[
            'carlton_file_path'
        ] = f"abfss://{config['container_src']}@{config['storage_name_src']}.dfs.core.windows.net/{config['path_src']}/"
        config[
            'schemaLocation'
        ] = f"abfss://{config['container_src']}@{config['storage_name_src']}.dfs.core.windows.net/_schemaLocation/{config['table_name']}/"
        config[
            'table_path'
        ] = f"abfss://{config['container_tgt']}@{config['storage_name_tgt']}.dfs.core.windows.net/{config['table_name']}/"
        config[
            'checkpointLocation'
        ] = f"abfss://{config['container_tgt']}@{config['storage_name_tgt']}.dfs.core.windows.net/{config['table_name']}/_checkpointLocation"
        return config
